import os
import pathlib
from os.path import join
from typing import TypeVar, Generic, Sequence

from canvasapi.exceptions import Unauthorized
from canvasapi.file import File
from canvasapi.folder import Folder

from .outputter import Outputter
from .utils import *

T = TypeVar('T')


class FileTree(Generic[T]):
    def __init__(self, path: str, folders: 'Sequence[FileTree[T]]', files: Sequence[T]):
        self.path = path
        self.folders: 'Sequence[FileTree[T]]' = folders
        self.files: Sequence[T] = files


def build_canvas_file_tree(base, folder: Folder) -> FileTree[File]:
    folders = list(map(lambda subfolder: build_canvas_file_tree(join(base, subfolder.name), subfolder), folder.list_folders()))
    files = list(folder.list_files())
    return FileTree(base, folders, files)


def length_file_tree(folder: FileTree) -> int:
    return sum(map(length_file_tree, folder.folders)) + len(folder.files)


class FileSynchronizer:
    def __init__(self, outputter):
        self.outputter = outputter

    def pull_file_tree(self, directory, tree):
        pathlib.Path(join(directory, tree.path)).mkdir(parents=True, exist_ok=True)

        for file in tree.files:
            filepath = join(join(directory, tree.path), file.filename)

            canvas_mtime = unix_time_seconds(file.modified_at_date.replace(tzinfo=pytz.utc))

            if not os.path.exists(filepath) or canvas_mtime > os.stat(filepath).st_mtime:
                file.download(filepath)
                atime = os.stat(filepath).st_atime
                os.utime(filepath, (atime, canvas_mtime))

        for subtree in tree.folders:
            self.pull_file_tree(directory, subtree)

    def pull_all_files(self, directory, course: Course):
        top_level_folder = next(folder for folder in course.list_folders() if folder.parent_folder_id is None)
        try:
            tree = build_canvas_file_tree('.', top_level_folder)
            self.outputter.poutput_verbose('Detected ' + str(length_file_tree(tree)) + ' files.')
            self.pull_file_tree(directory, tree)
        except Unauthorized:
            self.outputter.poutput('Not authorized to access this course\'s files')





import argparse
import functools

from .utils import get_course_by_query


def course_query_or_cc(clanvas, course, fail_on_ambiguous=False, quiet=False):
    if course is not None:
        return get_course_by_query(clanvas, course, fail_on_ambiguous=fail_on_ambiguous, quiet=quiet)
    elif clanvas.current_course is not None:
        return clanvas.current_course
    else:
        return None


def course_optional(argparser):
    argparser.add_argument('-c', '--course', default=None, help='course id or matching course code substring')
    return argparser


def argparser_course_optional_wrapper(with_argparser):
    """
    When applied to a do_x function in the Clanvas class that takes in argparser opts,
    will convert/replace course attribute with a corresponding course object either
    using the course string as a query or the current (cc'd) course if not supplied.
    :param with_argparser: the cmd2.with_argparser method to be wrapped.
    :return:
    """
    @functools.wraps(with_argparser)
    def inject_argparser(self, opts):
        opts.course = course_query_or_cc(self, opts.course)

        with_argparser(self, opts)

    return inject_argparser


DEFAULT = '__DEFAULT__'

cc_parser = argparse.ArgumentParser()
cc_parser.add_argument('course', nargs='?', default='',
                       help='course id or matching course string (e.g. the course code)')

cd_parser = argparse.ArgumentParser(description='Change the working directory.')
cd_parser.add_argument('directory', nargs='?', default='',
                       help='absolute or relative pathname of directory to become the new working directory')

lc_parser = argparse.ArgumentParser(description='List courses.')
lc_parser.add_argument('-a', '--all', action='store_true', help='all courses (previous terms)')
lc_parser.add_argument('-l', '--long', action='store_true', help='long listing')
lc_parser.add_argument('-i', '--invalidate', action='store_true', help='invalidate cached course info')

la_parser = argparse.ArgumentParser(description='List course assignments.')
la_parser = course_optional(la_parser)
la_parser.add_argument('-l', '--long', action='store_true', help='long listing')
la_parser.add_argument('-s', '--submissions', action='store_true', help='show submissions')
la_parser.add_argument('-u', '--upcoming', action='store_true', help='show only upcoming assignments')

lann_parser = argparse.ArgumentParser(description='List course announcements.')
lann_parser = course_optional(lann_parser)
lann_parser.add_argument('-n', '--number', type=int, default=None, help='number of announcements to display')
lann_parser.add_argument('-d', '--days', type=int, default=None, help='only show announcements this many days old')

catann_parser = argparse.ArgumentParser(description='Print course announcements.')
catann_parser = course_optional(catann_parser)
catann_parser.add_argument('ids', nargs='*', help='ids of announcements to print')

lg_parser = argparse.ArgumentParser(description='List course grades.')
lg_parser = course_optional(lg_parser)
lg_parser.add_argument('-l', '--long', action='store_true', help='long listing')
lg_parser.add_argument('-u', '--ungraded', action='store_false', help='include ungraded assignments')
lg_parser.add_argument('-a', '--all', action='store_true', help='show grades for all courses (previous terms) if no course is specified')

login_parser = argparse.ArgumentParser(description='Set URL and token to use for all Canvas API calls')
login_parser.add_argument('url', help='URL of Canvas server')
login_parser.add_argument('token', help='Canvas API access token')
login_parser.add_argument('-q', '--quiet', action='store_true', help='suppress login message')

pullf_parser = argparse.ArgumentParser(description='Pull course files to local disk.')
pullf_parser = course_optional(pullf_parser)
pullf_parser.add_argument('-o', '--output', help='location to save course files')

whoami_parser = argparse.ArgumentParser()
whoami_parser.add_argument('-v', '--verbose', action='store_true',
                           help='display more info about the logged in user')

wopen_parser = argparse.ArgumentParser(description='Open tabs in canvas web interface.')
wopen_parser = course_optional(wopen_parser)
wopen_parser.add_argument('tabs', nargs='*', default='', help='course tabs to open')


from abc import ABC
from enum import Enum


class Verbosity(Enum):
    OFF = 0
    NORMAL = 1
    VERBOSE = 2
    DEBUG = 3


class Outputter:
    def __init__(self, printfn, verbosityfn):
        self.printfn = printfn
        self.verbosityfn = verbosityfn

    def check(self, verbosity):
        return self.verbosityfn().value >= verbosity.value

    def poutput(self, msg, end='\n', verbosity=Verbosity.NORMAL):
        if self.check(verbosity):
            self.printfn(msg + end)

    def poutput_normal(self, msg, end='\n'):
        self.poutput(msg, end, verbosity=Verbosity.NORMAL)

    def poutput_verbose(self, msg, end='\n'):
        self.poutput(msg, end, verbosity=Verbosity.VERBOSE)

    def poutput_debug(self, msg, end='\n'):
        self.poutput(msg, end, verbosity=Verbosity.DEBUG)


from html2text import html2text

from .utils import *


class Printer:
    def __init__(self, outputter):
        self.outputter = outputter

    def print_announcement(self, course: Course, ids):
        if course is None:
            self.outputter.poutput('No course specified.')
            return False

        for announcement_id in ids:
            topic = course.get_discussion_topic(int(announcement_id))

            body = html2text('\n'.join(topic.message.splitlines())).strip()

            print_items = [topic.user_name,
                           long_datetime(topic.posted_at_date),
                           topic.title,
                           '',
                           body]

            self.outputter.poutput('\n'.join(print_items))

        return True



import os
import webbrowser
from os.path import isfile, join, expanduser
from urllib.parse import urlparse

import cmd2
import colorama
import readline
from canvasapi import Canvas
from colorama import Fore, Style

from .completers import Completers
from .filesynchronizer import FileSynchronizer
from .filters import latest_term_courses
from .interfaces import *
from .lister import Lister
from .outputter import Verbosity, Outputter
from .printer import Printer
from .utils import *


class Clanvas(cmd2.Cmd):
    default_to_shell = True

    def __init__(self):
        self.settable.update({'prompt_string': 'prompt format string'})
        self.settable.update({'verbosity': 'default command verbosity (NORMAL/VERBOSE/DEBUG)'})

        super(Clanvas, self).__init__()

        self.url = None
        self.host = None
        self.canvas = None  # type: Canvas

        self.home = os.path.expanduser("~")

        self.current_course = None  # type: Course
        self.current_directory = self.home

        self.outputter = Outputter(functools.partial(self.poutput, end=''), self.get_verbosity)

        self.file_synchronizer = FileSynchronizer(self.outputter)
        self.lister = Lister(self.outputter)
        self.printer = Printer(self.outputter)

        # Clanvas commands
        completers = Completers(self)
        for command, completer in completers.completer_mapping.items():
            setattr(self, 'complete_' + command, completer)
        # GNU commands
        for command in ['cat', 'tac', 'nl', 'od', 'base32', 'base64', 'fmt', 'tail', 'ls']:
            setattr(self, 'complete_' + command, functools.partialmethod(self.path_complete, dir_only=False))

    def get_caches(self):
        return self._caches

    @blocking_lru
    def get_courses(self, **kwargs):
        return {course.id: course for course in sorted(self.canvas.get_current_user().get_courses(include=['term', 'total_scores']),
                      key=lambda course: (-course.enrollment_term_id, course.name))}

    @blocking_lru
    def current_user_profile(self, **kwargs):
        return self.canvas.get_current_user().get_profile()

    @blocking_lru
    def list_tabs_cached(self, course_id):
        course = self.get_courses()[course_id]
        return sorted(course.list_tabs(), key=lambda tab: tab.position)

    @blocking_lru
    def list_announcements_cached(self, course_id):
        course = self.get_courses()[course_id]
        return sorted(course.get_discussion_topics(only_announcements=True), key=lambda t: t.posted_at_date)

    def get_verbosity(self) -> Verbosity:
        return Verbosity[self.verbosity]

    prompt_string = Fore.LIGHTGREEN_EX + '{login_id}@{host}' + Style.RESET_ALL + ':' + Fore.YELLOW + '{pwc}' + Style.RESET_ALL + ':' + Fore.BLUE + '{pwd} ' + Style.RESET_ALL + '$ '

    verbosity = 'NORMAL'

    canvas_path = expanduser('~/canvas')

    def get_prompt(self):
        if self.canvas is None:
            return '$ '

        return self.prompt_string.format(
            login_id=self.current_user_profile()['login_id'],
            host=self.host,
            pwc=self.current_course.course_code if self.current_course is not None else '~',
            pwd=os.getcwd().replace(self.home, '~')
        )


    # cmd2 attribute
    prompt = property(lambda self: self.get_prompt())

    #     _____                                          _
    #    / ____|                                        | |
    #   | |     ___  _ __ ___  _ __ ___   __ _ _ __   __| |___
    #   | |    / _ \| '_ ` _ \| '_ ` _ \ / _` | '_ \ / _` / __|
    #   | |___| (_) | | | | | | | | | | | (_| | | | | (_| \__ \
    #    \_____\___/|_| |_| |_|_| |_| |_|\__,_|_| |_|\__,_|___/
    #

    # Reimplement POSIX cd to call os.chdir

    @cmd2.with_argparser(cd_parser)
    def do_cd(self, opts):
        if opts.directory == '':
            path = os.path.expanduser('~')
        else:
            path = os.path.abspath(os.path.expanduser(opts.directory))

        if not os.path.isdir(path):
            self.poutput(f'cd: no such file or directory: {path}')
        elif not os.access(path, os.R_OK):
            self.poutput(f'cd: permission denied: {path}')
        else:
            try:
                os.chdir(path)
            except Exception as ex:
                self.poutput('{}'.format(ex))

    @cmd2.with_argparser(cc_parser)
    def do_cc(self, opts):
        if opts.course is '' or opts.course is '~':
            self.current_course = None
            return False

        match = get_course_by_query(self, opts.course)
        if match is not None:
            self.current_course = match

    @cmd2.with_argparser(lc_parser)
    def do_lc(self, opts):
        courses = self.get_courses().values()
        kwargs = dict(vars(opts))
        del kwargs['invalidate']
        self.lister.list_courses(courses, **kwargs)

    @cmd2.with_argparser(la_parser)
    @argparser_course_optional_wrapper
    def do_la(self, opts):
        return self.lister.list_assignments(**vars(opts))

    @cmd2.with_argparser(lg_parser)
    @argparser_course_optional_wrapper
    def do_lg(self, opts):
        opts_copy = dict(vars(opts))
        del opts_copy['all']
        if opts.course is None:
            del opts_copy['course']
            courses = self.get_courses().values() if opts.all else latest_term_courses(self.get_courses().values())
            for course in courses:
                self.lister.list_grades(course, **opts_copy)
        else:
            return self.lister.list_grades(**opts_copy)

    @cmd2.with_argparser(lann_parser)
    @argparser_course_optional_wrapper
    def do_lann(self, opts):
        return self.lister.list_announcements(**vars(opts), announcements_provider=self.list_announcements_cached)

    @cmd2.with_argparser(catann_parser)
    @argparser_course_optional_wrapper
    def do_catann(self, opts):
        return self.printer.print_announcement(**vars(opts))

    @cmd2.with_argparser(wopen_parser)
    @argparser_course_optional_wrapper
    def do_wopen(self, opts):
        if opts.course is None:
            self.poutput('No course specified.')
            return False

        course_tabs = self.list_tabs_cached(opts.course.id)
        given_tabs_set = frozenset([tab.lower() for tab in opts.tabs])

        matched_tabs = filter(lambda course_tab: course_tab.label.lower() in given_tabs_set, course_tabs)
        if not matched_tabs:
            self.poutput(f'No tab found matching "{tab}"')
            return False

        for tab in matched_tabs:
            webbrowser.open(tab.full_url, new=2)

        return True

    @cmd2.with_argparser(login_parser)
    def do_login(self, opts):
        if self.canvas is not None:
            self.poutput('Already logged in.')
            return False

        self.url = opts.url
        self.host = urlparse(opts.url).netloc

        self.canvas = Canvas(opts.url, opts.token)

        if not opts.quiet:
            profile = self.current_user_profile()
            self.poutput('Logged in as {:s} ({:s})'.format(profile['name'], profile['login_id']))
            call_eagerly(self.get_courses)
        else:
            call_eagerly(self.get_courses, self.current_user_profile)

    @cmd2.with_argparser(whoami_parser)
    def do_whoami(self, opts):
        profile = self.canvas.get_current_user().get_profile()

        if not opts.verbose:
            self.poutput(profile['name'] + ' (' + profile['login_id'] + ')')
        else:
            verbose_fields = ['name', 'short_name', 'login_id', 'primary_email', 'id', 'time_zone']
            self.poutput('\n'.join([field + ': ' + str(profile[field]) for field in verbose_fields]))

    @cmd2.with_argparser(pullf_parser)
    @argparser_course_optional_wrapper
    def do_pullf(self, opts):
        if opts.course is None:
            self.poutput('No course specified.')
            return False

        code = unique_course_code(opts.course)
        destination_path = join(*[os.path.expanduser('~'), 'canvas', 'courses', code, 'files']) if opts.output is None else opts.output

        self.file_synchronizer.pull_all_files(destination_path, opts.course)




def main():
    if 'libedit' in readline.__doc__:
        readline.parse_and_bind("bind -e")
        readline.parse_and_bind("bind '\t' rl_complete")

    colorama.init()  # Windows color support

    cmd = Clanvas()

    rc_file = join(expanduser('~'), '.clanvasrc')
    if isfile(rc_file):
        cmd.onecmd('load ' + rc_file)

    cmd.cmdloop()


if __name__ == "__main__":
    main()


import functools
import io
import shlex
from contextlib import redirect_stderr

from .interfaces import wopen_parser, course_query_or_cc, catann_parser
from .utils import unique_course_code, filter_courses


def complete_any(text, line, begidx, endidx, completers):
    for completer in completers:
        completions = completer(text, line, begidx, endidx)
        if completions:
            return completions
    return []


def parse_partial(argparser, line):
    try:
        stream = io.StringIO()
        with redirect_stderr(stream):
            opts, _ = argparser.parse_known_args(shlex.split(line)[1:])
            return opts
    except ValueError as e:
        if str(e) == 'No closing quotation':
            print(line + '\'')
            return parse_partial(argparser, line + '\'')
    except:
        import traceback
        traceback.print_exc()
        return None


class Completers:
    def __init__(self, clanvas):
        self.clanvas = clanvas
        self._course_complete_flags = dict.fromkeys(['-c', '--course'], self._course_completer)

        self.completer_mapping = {
            'catann': functools.partial(self._course_option_completer, all_else=self._catann_tab_completer),
            'cc': self._course_completer,
            'cd': functools.partial(self.clanvas.path_complete, dir_only=True),
            'la': self._course_option_completer,
            'lann': self._course_option_completer,
            'lg': self._course_option_completer,
            'pullf': self._pullf_tab_completer,
            'wopen': functools.partial(self._course_option_completer, all_else=self._wopen_tab_completer)
        }

    def _course_completer(self, text, line, begidx, endidx):
        return list(map(unique_course_code, filter_courses(self.clanvas.get_courses().values(), line[begidx:endidx])))

    def _catann_tab_completer(self, text, line, begidx, endidx):
        opts = parse_partial(catann_parser, line[0:endidx])
        if opts is None:
            return []

        opts.course = course_query_or_cc(self.clanvas, opts.course, fail_on_ambiguous=True, quiet=True)
        if opts.course is None:
            return []

        matching_string = '' if len(opts.ids) == 0 or line[endidx-1] == ' ' else opts.ids[-1]

        announcements = [str(ann.id) for ann in self.clanvas.list_announcements_cached(opts.course.id)]
        matched_announcements = filter(lambda ann: ann.startswith(matching_string), announcements)

        return list(matched_announcements)

    def _wopen_tab_completer(self, text, line, begidx, endidx):
        opts = parse_partial(wopen_parser, line[0:endidx])
        if opts is None:
            return []

        opts.course = course_query_or_cc(self.clanvas, opts.course, fail_on_ambiguous=True, quiet=True)
        if opts.course is None:
            return []

        matching_string = '' if len(opts.tabs) == 0 or line[endidx-1] == ' ' else opts.tabs[-1]

        tabs = self.clanvas.list_tabs_cached(opts.course.id)
        matched_tabs = filter(lambda tab: tab.label.lower().startswith(matching_string.lower()), tabs)
        return list(map(lambda tab: shlex.quote(tab.label.lower()), matched_tabs))

    def _course_option_completer(self, text, line, begidx, endidx, all_else=None):
        return self.clanvas.flag_based_complete(
            text, line, begidx, endidx, flag_dict=self._course_complete_flags, all_else=all_else)

    def _pullf_tab_completer(self, text, line, begidx, endidx):
        output_flags = dict.fromkeys(['-o', '--output'], functools.partial(self.clanvas.path_complete, dir_only=True))
        return self.clanvas.flag_based_complete(text, line, begidx, endidx, flag_dict={
            **self._course_complete_flags,
            **output_flags})



from datetime import datetime, timedelta

import pytz


def latest_term_courses(courses):
    latest_term_int = max(course.enrollment_term_id for course in courses)
    return filter(lambda course: course.enrollment_term_id == latest_term_int, courses)


def future_assignments(assignments):
    now = pytz.UTC.localize(datetime.now())
    return filter(lambda assignment: assignment.due_at_date >= now, assignments)


def days_from_today(iterable, days, key):
    target = pytz.UTC.localize(datetime.now() - timedelta(days=days))
    return filter(lambda item: key(item) >= target, iterable)


import os
import runpy
import sys

path = os.path.dirname(sys.modules[__name__].__file__)
path = os.path.join(path, '..')
sys.path.insert(0, path)
runpy.run_module('clanvas.clanvas', run_name="__main__",alter_sys=True)


import datetime
import functools
import threading
from collections import defaultdict

import pytz
from canvasapi.course import Course
from tabulate import tabulate
from tzlocal import get_localzone


def rstrip_zeroes(float_var):
    return ('%f' % float_var).rstrip('0').rstrip('.')


def compact_datetime(datetime_var):
    return datetime_var.astimezone(get_localzone()).strftime("%m-%d %I:%M%p")

def long_datetime(datetime_var):
    return datetime_var.astimezone(get_localzone()).strftime("%A, %B %d, %Y %I:%M%p")

def percentage_string(val, digits):
    val *= 10 ** (digits + 2)
    return '{1:.{0}f}%'.format(digits, val / 10 ** digits)


def rstripped_fraction(score, possible):
    numerator = rstrip_zeroes(score) if type(score) == float else str(score)
    denominator = rstrip_zeroes(possible) if type(possible) == float else str(possible)
    return f'{numerator}/{denominator}'


def unique_course_code(course):
    return course.course_code.replace(' ', '') + '-' + str(course.id)


def assignment_info_items(a):
    return [a.id, compact_datetime(a.due_at_date) if hasattr(a, 'due_at_date') else '', a.name]


def submission_info_items(s):
    return [s.id, s.score if hasattr(s, 'score') else '']


def filter_courses(courses, query):
    query_processed = query.replace(' ', '').lower()
    return filter(lambda course: query_processed in unique_course_code(course).replace(' ', '').lower(), courses)


def tabulate_dict(item_to_list, items):
    return dict(zip(items, tabulate(map(item_to_list, items), tablefmt='plain').split('\n')))


epoch = datetime.datetime.utcfromtimestamp(0).replace(tzinfo=pytz.utc)


def unix_time_seconds(dt):
    return (dt - epoch).total_seconds()


def get_course_by_query(clanvas, query, fail_on_ambiguous=False, quiet=False):
    matched_courses = list(filter_courses(clanvas.get_courses().values(), query))
    num_matches = len(matched_courses)

    if num_matches == 1:
        return matched_courses[0]
    elif num_matches > 1 and not fail_on_ambiguous:
        if not quiet:
            clanvas.poutput('Ambiguous course input "{:s}".'.format(query))
        if not quiet:
            clanvas.poutput('Please select an option:')

        pad_length = len(str(num_matches)) + 2
        format_str = '{:<' + str(pad_length) + '}{}'

        if not quiet:
            clanvas.poutput(format_str.format('0)', 'cancel'))
        count = 1
        for course in matched_courses:
            if not quiet:
                clanvas.poutput(format_str.format(f'{count})', unique_course_code(course)))
            count += 1

        choice = input('Enter number: ')
        if choice.isdigit():
            num_choice = int(choice)
            if num_choice > num_matches:
                if not quiet:
                    clanvas.poutput(f'Choice {num_choice} greater than last choice ({num_matches}).')
            elif num_choice != 0:
                return matched_courses[num_choice - 1]
        else:
            if not quiet:
                clanvas.poutput(f'Choice {choice} is not numeric.')
    else:
        if not quiet:
            clanvas.poutput('Could not find a matching course.' if not fail_on_ambiguous else 'Ambiguous course query string.')

    return None


def call_eagerly(*args):
    def runner(stop_event, callables):
        while not stop_event.is_set() and callables:
            callables.pop()()

    event = threading.Event()
    thread = threading.Thread(target=runner, args=(event, list(args)))
    thread.start()
    return event


def blocking_lru(func):
    func = functools.lru_cache(maxsize=None)(func)
    lock_dict = defaultdict(threading.Lock)

    def _thread_lru(*args, **kwargs):
        key = functools._make_key(args, kwargs, typed=False)
        with lock_dict[key]:
            return func(*args, **kwargs)

    return _thread_lru


def get_submissions_for_assignments(course: Course, assignments):
    assignment_ids = [assignment.id for assignment in assignments]
    assignment_submissions = course.list_multiple_submissions(assignment_ids=assignment_ids)

    submissions_by_assignment = defaultdict(list)

    for submission in assignment_submissions:
        submissions_by_assignment[submission.assignment_id].append(submission)

    return submissions_by_assignment



from setuptools import setup

setup(
    name='clanvas',
    version='0.2.0',
    url='https://github.com/MarkLalor/Clanvas',
    license='MIT',
    author='Mark Lalor',
    author_email='markwlalor@gmail.com',
    install_requires=['canvasapi', 'cmd2', 'tabulate', 'tree-format', 'html2text', 'colorama', 'pytz', 'tzlocal'],
    packages=['clanvas'],
    entry_points={'console_scripts': ['clanvas = clanvas.clanvas:main']},
    description='Command-line client for Canvas by Instructure',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Education',
        'Topic :: System :: Shells',
        'License :: OSI Approved :: MIT License'
    ]
)



from functools import lru_cache
from operator import itemgetter

from canvasapi.assignment import Assignment, AssignmentGroup
from canvasapi.exceptions import Unauthorized, CanvasException
from canvasapi.submission import Submission
from colorama import Fore, Style, Back
from html2text import html2text
from tree_format import format_tree

from .filters import latest_term_courses, future_assignments, days_from_today
from .outputter import Outputter
from .utils import *


def calculate_group_ratio(group, assignment_submission_pairs):
    # Memoize by sticking it in the group... maybe this is totally pointless though since arithmetic ops are cheap
    if hasattr(group, 'clanvas_total_points'):
        return group.clanvas_ratio, group.clanvas_total_points, group.clanvas_total_possible

    total_points = sum([(s.score if s is not None else 0) for (a, s) in assignment_submission_pairs])
    total_possible = sum([(a.points_possible if s is not None else 0) for (a, s) in assignment_submission_pairs])
    ratio = total_points / total_possible if total_possible != 0 else None
    group.clanvas_total_points = total_points
    group.clanvas_total_possible = total_possible
    group.clanvas_ratio = ratio
    return ratio, total_points, total_possible


class Lister:
    def __init__(self, outputter):
        self.outputter = outputter

    def list_courses(self, courses, all=False, long=False):
        display_courses = courses if all else latest_term_courses(courses)

        if long:
            def course_info_items(c):
                return [c.course_code, c.id, c.term['name'] if hasattr(c, 'term') else '', c.name]

            self.outputter.poutput(tabulate(map(course_info_items, display_courses), tablefmt='plain'))
        else:
            self.outputter.poutput('\n'.join([unique_course_code(c) for c in display_courses]))

    def list_assignments(self, course: Course, long=False, submissions=False, upcoming=False):
        if course is None:
            self.outputter.poutput('No course specified.')
            return False

        assignments = course.get_assignments()

        if upcoming:
            assignments = future_assignments(assignments)

        if long:
            if submissions:
                assignment_ids = map(lambda assignment: assignment.id, assignments)
                assignment_submissions = course.list_multiple_submissions(assignment_ids=assignment_ids)

                submissions_by_assignment = defaultdict(list)

                tabulated_submissions = tabulate_dict(submission_info_items, assignment_submissions)
                for submission, formatted in tabulated_submissions.items():
                    submissions_by_assignment[submission.assignment_id].append((formatted, []))

                tabulated_assignments = tabulate_dict(assignment_info_items, assignments)

                tree = (unique_course_code(course), [(formatted, submissions_by_assignment[assignment.id]) for assignment, formatted in tabulated_assignments.items()])

                self.outputter.poutput(format_tree(tree, format_node=itemgetter(0), get_children=itemgetter(1)))
            else:
                self.outputter.poutput(tabulate(map(assignment_info_items, assignments), tablefmt='plain'))
        else:
            self.outputter.poutput('\n'.join([assignment.name for assignment in assignments]))

    grade_color_thresholds = {
        0.9: Fore.LIGHTGREEN_EX,
        0.8: Fore.MAGENTA,
        0.7: Fore.YELLOW,
        0.6: Fore.LIGHTRED_EX,
        0.0001: Fore.RED,
        0.0: Back.RED + Fore.BLACK
    }

    @staticmethod
    def best_color(ratio):
        for key, value in Lister.grade_color_thresholds.items():
            if ratio >= key:
                return value

    @staticmethod
    def tabulate_grade_row(assignment, submission, long):
        possible = assignment.points_possible
        if submission is not None:
            score = submission.score
            fraction = rstripped_fraction(score, possible)
            if possible != 0:
                ratio = score / possible
                percentage = percentage_string(ratio, 0)
                color = Lister.best_color(ratio)
            else:
                percentage = 'N/A'
                color = ''

            def colored(string):
                return color + str(string) + Style.RESET_ALL
        else:
            fraction = rstripped_fraction('?', possible)
            percentage = ''

            def colored(string):
                return string

        if long:
            due_at_str = compact_datetime(assignment.due_at_date) if hasattr(assignment, 'due_at_date') else ''
            return [colored(string) for string in [submission.id if submission is not None else 'Unsubmitted', due_at_str, assignment.name, fraction, percentage]]
        else:
            return [colored(string) for string in [assignment.name, fraction, percentage]]

    @staticmethod
    def grades_tree(course: Course):
        assignment_submission_pair = course.get_assignments()
        submissions_by_assignment = get_submissions_for_assignments(course, assignment_submission_pair)

        def graded_submission(assignment):
            if assignment.id in submissions_by_assignment:
                submissions = submissions_by_assignment[assignment.id]
                submission = next((s for s in submissions if hasattr(s, 'grade') and s.grade is not None), None)
                if submission is not None:
                    return submission
            return None

        graded_assignment_submissions = {assignment: graded_submission(assignment) for assignment in assignment_submission_pair}

        tree_items = []
        assignment_groups = course.list_assignment_groups()

        grouped_assignment_submission_pairs = {}
        for assignment, submission in graded_assignment_submissions.items():
            if assignment.assignment_group_id not in grouped_assignment_submission_pairs:
                grouped_assignment_submission_pairs[assignment.assignment_group_id] = []
            grouped_assignment_submission_pairs[assignment.assignment_group_id].append((assignment, submission))

        for group in sorted(assignment_groups, key=lambda g: g.position):
            assignment_submission_pair = grouped_assignment_submission_pairs[group.id]\
                if group.id in grouped_assignment_submission_pairs else []
            group_subtree = (group, assignment_submission_pair)
            tree_items.append(group_subtree)

        return course, tree_items

    def list_grades(self, course: Course, long=False, ungraded=True):
        if course is None:
            self.outputter.poutput('No course specified.')
            return False

        try:
            tree = Lister.grades_tree(course)

            @lru_cache(maxsize=None)
            def group_ratio(id):
                pass

            def format_node(node):
                item = node[0]
                if isinstance(item, Course):
                    course = item
                    groups_item = node[1]

                    def weighted_contribution(group, assignment_submission_pairs):
                        ratio = calculate_group_ratio(group, assignment_submission_pairs)[0]
                        return (group.group_weight/100) * ratio if ratio is not None else group.group_weight/100

                    weighted_sum = sum(weighted_contribution(group, assignments) for (group, assignments) in groups_item)
                    if weighted_sum > 0:
                        percentage = percentage_string(weighted_sum, 1)
                        color = Lister.best_color(weighted_sum) if weighted_sum > 0 else ''
                        return color + course.name + ' ' + percentage + Style.RESET_ALL
                    else:
                        return course.name
                elif isinstance(item, AssignmentGroup):
                    group = item
                    assignment_submission_pairs = node[1]

                    ratio, total_points, total_possible = calculate_group_ratio(group, assignment_submission_pairs)

                    fraction = rstripped_fraction(total_points, total_possible)
                    if total_possible != 0:
                        ratio = total_points / total_possible
                        percentage = percentage_string(ratio, 1)
                        color = Lister.best_color(ratio)
                    else:
                        percentage = 'N/A'
                        color = ''

                    if long:
                        items = [group.name, f'(Weight={percentage_string(group.group_weight/100, 0)})', fraction, percentage]
                    else:
                        items = [group.name, fraction, percentage]

                    return color + ' '.join(items) + Style.RESET_ALL
                elif isinstance(item, Assignment):
                    assignment, submission = node
                    return ' '.join([str(x) for x in Lister.tabulate_grade_row(assignment, submission, long=long)])

            def get_children(node):
                if isinstance(node[1], Submission) or node[1] is None:
                    return []
                else:
                    return list(filter(lambda item: ungraded or item[1] is not None, node[1]))

            self.outputter.poutput(format_tree(tree, format_node=format_node, get_children=get_children), end='')
        except Unauthorized:
            self.outputter.poutput(f'{course.name}: Unauthorized')
        except CanvasException as e:
            self.outputter.poutput(f'{course.name}: {str(e)}')

    def list_announcements(self, announcements_provider, course: Course, number=None, days=None, message=False):
        if course is None:
            self.outputter.poutput('No course specified.')
            return False

        display_topics = announcements_provider(course.id)

        if number is not None:
            display_topics = display_topics[-number:]

        if days is not None:
            display_topics = days_from_today(display_topics, days, key=lambda t: t.posted_at_date)


        if message:
            def print_topic(topic):
                return '\n'.join([topic.user_name, topic.title, html2text('\n'.join(topic.message.splitlines())).strip()])
            output = '\n=================\n'.join(map(print_topic, display_topics))
        else:
            def topic_row(topic):
                return [topic.id, compact_datetime(topic.posted_at_date), topic.user_name, topic.title]
            output = tabulate(map(topic_row, display_topics), tablefmt='plain')

        self.outputter.poutput(output)

        return False
